# -*- coding: utf-8 -*-

"""Main module."""

from collections import abc, namedtuple
import itertools
from uuid import uuid4


Arc = namedtuple('Arc', 'initial_node terminal_node')


class Graph():
    def __init__(self, nodes=[], arcs=[]):
        if not isinstance(nodes, abc.Iterable):
            raise TypeError('expected nodes to be iterable')

        if not isinstance(arcs, abc.Iterable):
            raise TypeError('expected arcs to be iterable')

        def neighbors_of(node):
            '''Return the multiset of neighbors of node based on the argument
            passed to arcs.
            '''

            return tuple(arc[1] for arc in arcs if arc[0] == node)

        nodes_from_arcs = {node for arc in arcs for node in arc}
        all_nodes = set(nodes).union(nodes_from_arcs)
        self._dict = {node: neighbors_of(node) for node in all_nodes}

    def __repr__(self):
        '''Return a string representation of the graph (for debugging).'''

        name = self.__class__.__name__
        arcs = [tuple(arc) for arc in self.arcs]
        nodes = list(self.nodes)
        s = '{}(nodes={}, arcs={})'.format(name, nodes, arcs)
        return s

    def __str__(self):
        '''Return a string representation of the graph (for printing) as an
        adjacency list.'''

        return str(self._dict)

    def __getitem__(self, node):
        '''Return the set of neighbor of node'''

        if node not in self._dict:
            raise KeyError('graph has no node {!r}'.format(node))

        return set(self._dict[node])

    @property
    def nodes(self):
        '''Return the set of nodes in the graph.'''

        return set(node for node in self._dict)

    @property
    def arcs(self):
        '''Return the multiset of arcs in the graph.'''

        arcs = [
            Arc(initial_node=node, terminal_node=neighbor)
            for node in self._dict
            for neighbor in self._dict[node] if len(self._dict[node]) > 0
        ]
        return arcs

    @property
    def order(self):
        '''Return the number of nodes in the graph.'''

        return len(self._dict)

    @property
    def size(self):
        '''Return the number of arcs in the graph.'''

        return sum(len(self._dict[node]) for node in self._dict)

    def neighbors_of(self, node):
        '''Return the set of neighbor of node.'''

        return self.__getitem__(node)

    def has_nodes(self, nodes):
        '''Return True if every node in nodes is a node in the graph and
        False otherwise.'''

        if not isinstance(nodes, abc.Iterable):
            raise TypeError('nodes should be iterable')

        for node in nodes:
            if node not in self.nodes:
                return False
        return True

    def has_path(self, start=None, end=None):
        '''Return true if the graph has a path from start to end.'''

        valid_start_and_end = self.has_nodes([start, end])
        if not valid_start_and_end:
            return False

        if start == end:
            return True

        if end in self.neighbors_of(start):
            return True

        for node in self.neighbors_of(start):
            if self.has_path(start=node, end=end):
                return True
        return False

    @property
    def is_strongly_connected(self):
        '''Return True if the graph is strongly connected and False
        otherwise.'''

        for pair_of_nodes in itertools.combinations(self.nodes, 2):
            start, end = pair_of_nodes
            if not self.has_path(start=start, end=end):
                return False
        return True

    def degree_of(self, node):
        '''Return the number of neighbors of node in the graph.'''

        number_of_neighbors = len(self.neighbors_of(node))
        return number_of_neighbors


def read_graph_from_edge_list(filepath, delim=' '):
    '''Return an instance of the Graph class from an edge list located at
    filepath.'''

    with open(filepath, 'r') as f:
        lines = (line.strip() for line in f)
        arcs = [line.split(delim) for line in lines]
        return Graph(arcs=arcs)


def write_edge_list_from_graph(graph, filename=None):
    '''Write graph data to file as edge list. Returns the filename of the
    output file.

    Note that only nodes contained in arcs will be written, so writing from
    graphs with singletons will result in a loss of data.'''

    if not filename:
        uuid = uuid4()
        filename = 'Graph{}'.format(uuid)

    with open(filename, 'w') as f:
        lines = ('{} {}'.format(*arc) for arc in graph.arcs)
        out_string = '\n'.join(lines)
        f.write(out_string)
        return filename
